import * as THREE from 'three';

export declare type CSMPatchMap = {
    [keyword: string]: {
        [toReplace: string]: string | {
            value: string;
            type: "fs" | "vs";
        };
    };
};

export declare type CSMProxy<T extends MaterialConstructor> = InstanceType<T> & {
    vertexShader: string;
    fragmentShader: string;
    uniforms: Uniform;
    update: (opts: Omit<CustomShaderMaterialBaseParameters<T>, "baseMaterial">) => void;
    __csm: {
        prevOnBeforeCompile: THREE.Material["onBeforeCompile"];
    };
};

declare class CustomShaderMaterial<T extends TYPES.MaterialConstructor = typeof THREE.Material> extends THREE.Material {
    uniforms: TYPES.Uniform;
    vertexShader: string;
    fragmentShader: string;
    constructor({ baseMaterial, vertexShader, fragmentShader, uniforms, patchMap, cacheKey, ...opts }: TYPES.CustomShaderMaterialParameters<T>);
    update({ fragmentShader: _fs, vertexShader: _vs, uniforms, cacheKey, patchMap, }: Omit<TYPES.CustomShaderMaterialBaseParameters<T>, "baseMaterial">): void;
}
export default CustomShaderMaterial;

declare type CustomShaderMaterialBaseParameters<T extends MaterialConstructor> = {
    baseMaterial: T | InstanceType<T>;
    vertexShader?: string;
    fragmentShader?: string;
    uniforms?: Uniform;
    patchMap?: CSMPatchMap;
    cacheKey?: () => string;
};

export declare type CustomShaderMaterialParameters<T extends MaterialConstructor> = CustomShaderMaterialBaseParameters<T> & (MaterialParams<T> extends undefined ? any : MaterialParams<T>);

declare interface IUniform<TValue = any> {
    value: TValue;
}

export declare type MaterialConstructor = new (...opts: any[]) => THREE.Material;

declare type MaterialParams<T extends MaterialConstructor> = ConstructorParameters<T>[0];

declare namespace TYPES {
    export {
        IUniform,
        Uniform,
        MaterialConstructor,
        CSMPatchMap,
        CustomShaderMaterialBaseParameters,
        CustomShaderMaterialParameters,
        CSMProxy
    }
}

declare type Uniform = {
    [key: string]: IUniform;
};

export { }
